﻿# CODE BLOCK BEGIN Global_Definitions 
from TSMaster import *
from Crypto.Hash import CMAC    # CMAC 계산에 필요하다.
from Crypto.Cipher import AES   # CMAC 계산의 알고리즘이다.
import random                   # 난수로 임의의 data를 만든다. 

# m_id
k_m_id_secoc_coord = 0x100
k_m_id_ecu_1 = 0x101
k_m_id_ecu_2 = 0x102

# CMAC을 사용한다.
# key는 128 비트, 16 바이트이다.
# 이 키는 비밀로 간직해야 한다.
# 실제로는 이렇게 코드에 노출하지 않는다.
k_key =  bytearray([0x00, 0x01, 0x02, 0x03, 0x04, 0x5, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f])

# 키 길이
k_len_key = len(k_key)  # Length of key

# CMAC 길이. 128 비트, 16 "바이트"다.
k_len_cmac = 16 # Length of CMAC

# CMAC_trunc의 길이
# CMAC 16 바이트 중에서 앞 2 바이트를 사용한다.
k_len_cmac_trunc = 2 # Length of CMAC truncated

# Freshness Value의 길이
# fv를 6 바이트로 한다.  
k_len_fv = 6 # Length of fv

# fv_trunc의 길이
# fv 6 바이트 중에 1 바이트를 사용한다.
k_len_fv_trunc = 1 # Length of fv truncated

# MAC(Message Authentication Code)의 길이
# fv_trunc 1 바이트와 cmac_trunc 2 바이트 해서 3 바이트이다.
k_len_mac = k_len_fv_trunc + k_len_cmac_trunc # Length of MAC

# CAN/-FD의 최대 메시지 길이
# python은 메지지 길이에 맞게 동적으로 저장 공간을 확보할 수 있지만
# TSMaster를 보니, 실제 메시지 길이와 무관하게 
# 모든 메시지에 64 바이트 저장 공간을 확보한다. 
# 이렇게 하는 것이 어떤 이점이 있는지 알아보려고 따라해본다. 
k_len_data_max = 64 # Maximum data length

# fv의 최대값
# fv가 6 바이트이면, fv의 최대값은 0xFF가 6개 이다.
k_fv_max = int(bytes([0xFF] * k_len_fv).hex(), 16)   

# rx 메시지의 fv를 관리하기 위한 딕셔너리
rx_msgs = {}    

# tx 메시지의 fv를 관리하기 위한 딕셔너리
tx_msgs = {}    

# tx 메시지의 fv를 미리 정해둔다.
tx_msgs[k_m_id_ecu_1] = {
    'fv': 0,
    # fv외에 다른 데이터를 저장할 수 있도록 딕셔너리로 선언한다.
}

tx_msgs[k_m_id_ecu_2] = {
    'fv': bytearray([0x00] * k_len_fv)
}

# ecu_1이 전송하는 메시지를 정의한다.
# m_id, dlc, channel, 0, data, 설정 순서다. 
msg_ecu_1 = RawCAN(
    k_m_id_ecu_1, 
    0x0a, 
    CH1, 
    0, 
    bytearray([i for i in range(16)]), 
    0x306
)
 
# MAC(= fv_trunc, cmac_trunc)을 메시지의 마지막에 붙일 것이다. 
# 데이터의 끝에서 fv_trunc의  위치
k_pos_fv_trunc_from_eod = k_len_mac   

# 데이터의 끝에서 cmac_trunc의 위치
k_pos_cmac_trunc_from_eod = k_len_cmac_trunc 


# function
def dlc_to_length(dlc, is_can_fd):
    '''
    아래 표에 따라 dlc(data length code) 값을 길이로 변환한다.
    CAN의 경우 dlc는 길이이다.
    CAN-FD의 경우, dlc는 길이가 아니다.
    
    dlc can can-fd
    0	0	0
    1	1	1
    2	2	2
    3	3	3
    4	4	4
    5	5	5
    6	6	6
    7	7	7
    8	8	8
    9	8	12
    10	8	16
    11	8	20
    12	8	24
    13	8	32
    14	8	48
    15	8	64
    '''

    length = [0, 1, 2, 3, 4, 5, 6, 7, 8, 12, 16, 20, 24, 32, 48, 64]

    if is_can_fd:
        return length[dlc]
    else:
        return min(length[dlc], 8)


def check_fv(ACAN):
    '''
    fv가 맞는지 확인한다.
        수신한 fv_trunc와 계산한 fv_trunc_calc가동일하면
        fv가 맞다고 간주한다.   
    ACAN에서 fv_trunc를 추출한다.
        fv_trunc
            ACAN의 data에 있다. 
            fv_trunc는 k_len_fv_trunc 바이트이다. 
            pos_fv는 data 내에 fv_trunc의 위치이다.
                    pos_fv = m_len - k_fv_pos_from_eod
                           m_len은 메시지 길이이다.
    트래킹 중인 fv에서 fv_trnc_calc를 추출한다.
        fv_trunc_calc는 fv의 마지막 바이트이다.
    fv_trunc와 fv_trunc_calc를 비교한다.
    '''

    m_id = ACAN.id
    m_len = dlc_to_length(ACAN.dlc, ACAN.is_edl)

    ## fv
    # 트랙깅하는 fv를 1만큼 증가한다.
    # cmac 확인까지 마친 후에 fv를 증가시켜야 한다. 
    # 확인 전까지 fv_copy를 이용한다.
    fv_copy = rx_msgs[m_id]['fv']  
    fv_copy += 1

    # fv_copy가 k_fv_max를 넘어가면 0으로 초기화한다.
    # if fv_copy > 0xFF FF FF FF FF FF:
    if fv_copy > k_fv_max:
        fv_int = 0

    # fv_trunc_calc를 구한다.
    # int를 k_len_fv 길이의 bytearray로 변환하고
    # 마지막 바이트를 취한다.
    # 1 바이트는 헥스로 2자리라서 2를 곱한다. 
    fv_trunc_calc = bytearray.fromhex(f'{fv_copy:0{k_len_fv * 2}x}')[-1]

    # fv_trunc를 구한다.
    # fv_trunc의 위치
    pos_fv_trunc = m_len - k_pos_fv_trunc_from_eod

    # fv_trunc
    fv_trunc = ACAN.data[pos_fv_trunc]

    # moment of truth
    if fv_trunc != fv_trunc_calc:
        app.log_text(f'fv error: {m_id=:-2x} {fv_trunc=} {fv_trunc_calc=:}', lvlOK)
        return False, fv_copy
    
    return True, fv_copy


def check_cmac(ACAN, fv_copy):
    '''
    cmac이 맞는지 확인한다.
          수신한 cmac_trunc와 계산한 cmac_trunc_calc가 동일하면
          cmac이 맞다고 간주한다.
    ACAN에서 cmac_trunc를 추출한다.
        cmac_trunc
            ACAN의 data에 있다. 
            길이는 k_len_cmac_trunc 바이트이다.
            시작 위치는 m_len-k_pos_cmac_truc_from이다.
    k_key, data, fv(fv_trunc가 아니다.)를 이용하여 cmac을 계산한다.
    cmac에서 cmac_trunc_calc를 추출한다.
        cmac_trunc_calc는 cmac[:k_len_cmac_truc]이다.
        상위에서 k_len_cmac_trunc 만큼이다.
    cmac_trunc와 cmac_trunc_calc를 비교한다.
    '''

    m_id = ACAN.id
    m_len = dlc_to_length(ACAN.dlc, ACAN.is_edl)
    data = bytearray(ACAN.data[:m_len - k_len_mac])
    fv_bytearray = bytearray.fromhex(f'{fv_copy:0{k_len_fv * 2}x}')

    # 수신 메시지에서 cmac_trunc를 추출한다.
    pos_cmac_trunc = m_len - k_pos_cmac_trunc_from_eod
    cmac_trunc = bytearray(ACAN.data[pos_cmac_trunc:pos_cmac_trunc + k_len_cmac_trunc])

    # 수신 메시지의 데이터로 cmac을 계산한다.
    cmac_obj = CMAC.new(k_key, ciphermod=AES)
    cmac_obj.update(data)
    cmac_obj.update(fv_bytearray)
    cmac = cmac_obj.digest()

    # 계산한 cmac에서 cmac_trunc_calc를 추출한다..    
    cmac_trunc_calc = cmac[:k_len_cmac_trunc]

    # moment of truth
    if cmac_trunc != cmac_trunc_calc:
        app.log_text(f'cmac error: {cmac_trunc.hex()=} {cmac_trunc_calc.hex()=}', lvlOK)
        return False, cmac
    
    return True, cmac


def process_secoc_message(ACAN):
    '''
    fv가 맞는지 확인한다. 
    cmac이 맞는지 확인한다. 
    cmac이 맞으면, 트래킹하는 fv를 업데이트한다. 
    '''

    # fv가 맞는지 확인한다.
    fv_ok, fv_copy = check_fv(ACAN)
    if not fv_ok:  
        return
    
    # cmac이 맞는지 확인한다.
    cmac_ok, cmac = check_cmac(ACAN, fv_copy)
    if not cmac_ok:
        return

    # fv와 cmac이 맞으면, fv를 업데이트한다.
    m_id = ACAN.id
    rx_msgs[m_id]['fv'] = fv_copy
    app.log_text('ok', lvlOK)
    

def process_secoc_coord_message(ACAN):
    '''
    coord message를 처리한다.

    coord message의 용도
    1. fv 재설정을 위한 fv_init을 전달한다.

    coord message의 구조
    request + m_id + data
    request: 
        1 byte. 256개의 명령이 가능하다. 
        이렇게 많은 명령이 필요하지는 않을 것이다. 
        따라서 1 바이트로 충분하다.
        0x01: transfer fv_init

    m_id_target:
        2 bytes.
        대상 메시지를 지정한다.
        m_id가 0xAA 0xAA이면 모든 메시지들을 대상으로 하느 것으로 하자. 

    data: 
        secoc_coord 메시지는 16 바이트로 한다.
        request와 m_id를 제외하면 13 바이트가 남는다.
        데이터 길이는 request에 따라 다르다.
        request 0x01: fv_init
            data[3:3 + k_len_fv]는 fv_init이다. 
            m_id 메시지의 fv를 fv_init으로 초기화한다. 
            즉, rx_msgs[m_id]['fv'] = fv_init
    '''

    # 여기에 secoc_coord 메시지의 
    # cmac_trunc를 확인하는 과정이 필요하다.


    # 대상 메시지의 m_id를 추출한다.
    m_id_target = bytearray(ACAN.data[1:3])
    request = ACAN.data[0]

    if request == 0x01: # fv를 업데이트 한다.
        fv_bytearray = bytearray(ACAN.data[3:9]) 
        app.log_text(f'fv update m_id=0x{m_id_target.hex()} fv=0x{fv_bytearray.hex()}', lvlOK)
        if m_id_target == bytearray([0xAA, 0xAA]): # 모든 메시지들을 대상으로 한다.
            for msg in rx_msgs.values():
                msg['fv'] = int.from_bytes(fv_bytearray, byteorder='big')
                app.log_text(f'fv updated {msg=}', lvlOK)
        else:
            m_id_target_int = int(m_id_target.hex(), 16)
            rx_msgs[m_id_target_int]['fv'] = int.from_bytes(fv_bytearray, byteorder='big') 
            app.log_text(f'fv updated', lvlOK)

# CODE BLOCK END Global_Definitions 
# CODE BLOCK BEGIN Instance 
Instance = MpInstance('SecOC_ecu_1')
# CODE BLOCK END Instance 
# CODE BLOCK BEGIN On_CAN_FD_Rx any_messages LTEsLTEsLTE_
def on_canfd_rx_any_messages(ACAN: RawCAN) -> None:
    global rx_msgs

    m_id = ACAN.id

    if m_id not in rx_msgs.keys():        

        # add ACAN.id to msgs
        rx_msgs[m_id] = {
            'fv': 0,  
        }

        app.log_text(f'{m_id=:03x} added to rx_msgs {rx_msgs[m_id]}', lvlOK)

    if (m_id == k_m_id_secoc_coord):
        process_secoc_coord_message(ACAN)
    else:
        process_secoc_message(ACAN)

# CODE BLOCK END On_CAN_FD_Rx any_messages
# CODE BLOCK BEGIN Timer timer_1 NTAw
timer_1 = MpTimer("SecOC_ecu_1", "timer_1", 500) # Auto generated code, for modification please goto python code editor
# CODE BLOCK END Timer timer_1
# CODE BLOCK BEGIN On_Timer timier_1 dGltZXJfMQ__
def on_timer_timier_1() -> None:
    global msg_ecu_1
    global tx_msgs


    msg = msg_ecu_1
    msg.data = bytearray(random.randbytes(16))
    m_id = msg.id

#     app.log_text(f'{m_id = :02x}', lvlOK)
#     app.log_text(msg_ecu_1, lvlOK)
    
    m_len = tx_msgs[m_id]['len']
    
    # fv_trunc를 추가한다.
    fv = tx_msgs[m_id]['fv']
    fv_int = int(fv.hex(), 16) + 1

    if fv_int > 0xFFFFFFFFFFFF:
        fv_int = 0
    fv = bytearray.fromhex(f'{fv_int:012x}')
    fv_trunc = fv[-1]


#     app.log_text(f'{m_id = :0x} {m_len = }', lvlOK)
#     app.log_text(f'{fv = }', lvlOK)
#     app.log_text(f'{fv_int = :,}', lvlOK)
#     app.log_text(f'{fv_trunc = :02x}', lvlOK)

    pos_fv_trunc = m_len - k_pos_fv_trunc_from_eod

    msg.set_data(pos_fv_trunc, fv_trunc)

    cmac_obj = CMAC.new(k_key, ciphermod=AES)
    cmac_obj.update(msg.data[:m_len - k_len_mac])
    cmac_obj.update(fv)
    cmac = cmac_obj.digest()
    cmac_trunc = cmac[:k_len_cmac_trunc]

    pos_cmac_trunc = m_len - k_pos_cmac_trunc_from_eod
    msg.set_data(pos_cmac_trunc, cmac_trunc[0])
    msg.set_data(pos_cmac_trunc + 1, cmac_trunc[1])

    tx_msgs[m_id]['fv'] = fv
    tx_msgs[m_id]['cmac'] = cmac
 
    if not com.transmit_can_async(msg_ecu_1):
        app.log_text('transmit failed', lvlInfo)
    pass

# CODE BLOCK END On_Timer timier_1
# CODE BLOCK BEGIN On_Shortcut ctrl_1
def on_shortcut_ctrl_1() -> None:
    timer_1.Start()

# CODE BLOCK END On_Shortcut ctrl_1
# CODE BLOCK BEGIN On_Shortcut ctrl_0
def on_shortcut_ctrl_0() -> None:
    timer_1.Stop()

# CODE BLOCK END On_Shortcut ctrl_0
# CODE BLOCK BEGIN Step_Function  MA__
def step() -> None:
    pass

# CODE BLOCK END Step_Function 
